from __future__ import absolute_import

import functools
import time
import typing as tp
from struct import pack

# noinspection PyProtectedMember
from . import logger
from .utils import general_utils
from .errors import OutOfBoundsError
from .geometry import Region

if tp.TYPE_CHECKING:
    from .eyes_base import EyesBase
    from .utils.custom_types import (Num, RunningSession, AppOutput,
                                     UserInputs, MatchResult)
    from .agent_connector import AgentConnector
    from .match import ImageMatchSettings
    from .capture import EyesScreenshot

__all__ = ('MatchWindowTask',)


# TODO: remove Eyes and Target dependencies from here

class MatchWindowTask(object):
    """
    Handles matching of output with the expected output (including retry and 'ignore mismatch' when needed).
    """
    _MATCH_INTERVAL = 0.5

    MINIMUM_MATCH_TIMEOUT = 60  # Milliseconds

    def __init__(self, eyes, agent_connector, running_session, default_retry_timeout):
        # type: (EyesBase, AgentConnector, RunningSession, Num) -> None
        """
        Ctor.

        :param eyes: The Eyes instance which created this task.
        :param agent_connector: The agent connector to use for communication.
        :param running_session:  The current eyes session.
        :param default_retry_timeout: The default match timeout. (milliseconds)
        """
        self._eyes = eyes
        self._agent_connector = agent_connector
        self._running_session = running_session
        self._default_retry_timeout = default_retry_timeout / 1000.0  # type: Num # since we want the time in seconds.
        self._screenshot = None  # type: tp.Optional[EyesScreenshot]

    @staticmethod
    def _create_match_data_bytes(app_output,  # type: AppOutput
                                 user_inputs,  # type: UserInputs
                                 tag,  # type: tp.Text
                                 ignore_mismatch,  # type: bool
                                 screenshot,  # type: EyesScreenshot
                                 default_match_settings,  # type: ImageMatchSettings
                                 target,
                                 ignore=None,  # type: tp.Optional[tp.List]
                                 floating=None,  # type: tp.Optional[tp.List]
                                 ):
        # type: (...) -> bytes
        if ignore is None:
            ignore = []
        if floating is None:
            floating = []

        match_data = {
            "IgnoreMismatch": ignore_mismatch,
            "Options":        {
                "Name":               tag,
                "UserInputs":         user_inputs,
                "ImageMatchSettings": {
                    "MatchLevel":  default_match_settings.match_level,
                    "IgnoreCaret": target.ignore_caret,
                    "Exact":       default_match_settings.exact_settings,
                    "Ignore":      ignore,
                    "Floating":    floating
                },
                "IgnoreMismatch":     ignore_mismatch,
                "Trim":               {
                    "Enabled": False
                }
            },
            "UserInputs":     user_inputs,
            "AppOutput":      app_output,
            "tag":            tag
        }
        match_data_json_bytes = general_utils.to_json(match_data).encode('utf-8')  # type: bytes
        match_data_size_bytes = pack(">L", len(match_data_json_bytes))  # type: bytes
        screenshot_bytes = screenshot.get_bytes()
        body = match_data_size_bytes + match_data_json_bytes + screenshot_bytes
        return body

    @staticmethod
    def _get_dynamic_regions(target, eyes_screenshot):
        ignore = []  # type: tp.List[Region]
        floating = []  # type: tp.List[Region]
        if target is not None:
            for region_wrapper in target._ignore_regions:
                try:
                    current_region = region_wrapper.get_region(eyes_screenshot)
                    ignore.append(current_region)
                except OutOfBoundsError as err:
                    logger.info("WARNING: Region specified by {} is out of bounds! {}".format(region_wrapper, err))
            for floating_wrapper in target._floating_regions:
                try:
                    current_floating = floating_wrapper.get_region(eyes_screenshot)
                    floating.append(current_floating)
                except OutOfBoundsError as err:
                    logger.info("WARNING: Floating region specified by {} is out of bounds! {}".format(floating_wrapper,
                                                                                                       err))
        return {"ignore": ignore, "floating": floating}

    def _prepare_match_data_for_window(self, tag,  # type: tp.Text
                                       user_inputs,  # type: UserInputs
                                       default_match_settings,  # type: ImageMatchSettings
                                       target,
                                       ignore_mismatch=False):
        # type: (...) -> bytes
        title = self._eyes._title
        # TODO: Refactor this
        if hasattr(self._eyes, '_hide_scrollbars_if_needed'):
            with self._eyes._hide_scrollbars_if_needed():  # type: ignore
                self._screenshot = self._eyes.get_screenshot(hide_scrollbars_called=True)
        else:
            self._screenshot = self._eyes.get_screenshot()

        dynamic_regions = MatchWindowTask._get_dynamic_regions(target, self._screenshot)
        app_output = {'_title': title, 'screenshot64': None}  # type: AppOutput

        if self._eyes.send_dom:
            dom_json = self._eyes._try_capture_dom()
            if dom_json:
                dom_url = self._eyes._try_post_dom_snapshot(dom_json)
                if dom_url is None:
                    logger.warning('Failed to upload DOM. Skipping...')
                else:
                    app_output['DomUrl'] = dom_url

        logger.debug('AppOutput: {}'.format(app_output))
        return self._create_match_data_bytes(app_output, user_inputs, tag, ignore_mismatch,
                                             self._screenshot, default_match_settings, target,
                                             dynamic_regions['ignore'], dynamic_regions['floating'])

    def _run_with_intervals(self, prepare_action, retry_timeout):
        # type: (tp.Callable, Num) -> MatchResult
        """
        Includes retries in case the screenshot does not match.
        """
        logger.debug('Matching with intervals...')
        # We intentionally take the first screenshot before starting the timer, to allow the page
        # just a tad more time to stabilize.
        data = prepare_action()
        # Start the timer.
        start = time.time()
        logger.debug('First match attempt...')
        as_expected = self._agent_connector.match_window(self._running_session, data)
        if as_expected:
            return {"as_expected": True, "screenshot": self._screenshot}
        retry = time.time() - start
        logger.debug("Failed. Elapsed time: {0:.1f} seconds".format(retry))

        while retry < retry_timeout:
            logger.debug('Matching...')
            time.sleep(self._MATCH_INTERVAL)
            data = prepare_action()
            as_expected = self._agent_connector.match_window(self._running_session, data)
            if as_expected:
                return {"as_expected": True, "screenshot": self._screenshot}
            retry = time.time() - start
            logger.debug("Elapsed time: {0:.1f} seconds".format(retry))
        # One last try
        logger.debug('One last matching attempt...')
        data = prepare_action()
        as_expected = self._agent_connector.match_window(self._running_session, data)
        return {"as_expected": as_expected, "screenshot": self._screenshot}

    def _run(self, prepare_action, run_once_after_wait=False, retry_timeout=-1):
        # type: (tp.Callable, bool, Num) -> MatchResult
        if 0 < retry_timeout < MatchWindowTask.MINIMUM_MATCH_TIMEOUT:
            raise ValueError("Match timeout must be at least 60ms, got {} instead.".format(retry_timeout))
        if retry_timeout < 0:
            retry_timeout = self._default_retry_timeout
        else:
            retry_timeout /= 1000.0
        logger.debug("Match timeout set to: {0} seconds".format(retry_timeout))
        start = time.time()
        if run_once_after_wait or retry_timeout == 0:
            logger.debug("Matching once...")
            # If the load time is 0, the sleep would immediately return anyway.
            time.sleep(retry_timeout)
            data = prepare_action()
            as_expected = self._agent_connector.match_window(self._running_session, data)
            result = {"as_expected": as_expected, "screenshot": self._screenshot}  # type: MatchResult
        else:
            result = self._run_with_intervals(prepare_action, retry_timeout)
        logger.debug("Match result: {0}".format(result["as_expected"]))
        elapsed_time = time.time() - start
        logger.debug("_run(): Completed in {0:.1f} seconds".format(elapsed_time))
        return result

    def match_window(self, retry_timeout,  # type: Num
                     tag,  # type: str
                     user_inputs,  # UserInputs
                     default_match_settings,  # type: ImageMatchSettings
                     target,
                     ignore_mismatch,
                     run_once_after_wait=False):
        # type: (...) -> MatchResult
        """
        Performs a match for the window.

        :param retry_timeout: Amount of time until it retries.
        :param tag: The name of the tag (optional).
        :param user_inputs: The user input.
        :param default_match_settings: The default match settings for the session.
        :param target: The target of the check_window call.
        :param ignore_mismatch: True if the server should ignore a negative result for the visual validation.
        :param run_once_after_wait: Whether or not to run again after waiting.
        :return: The result of the run.
        """
        prepare_action = functools.partial(self._prepare_match_data_for_window, tag,
                                           user_inputs, default_match_settings, target, ignore_mismatch)
        return self._run(prepare_action, run_once_after_wait, retry_timeout)



import typing as tp

from .errors import EyesError

__all__ = ('MatchLevel', 'ExactMatchSettings', 'ImageMatchSettings')


class MatchLevel(object):
    """
    The extent in which two images match (or are expected to match).
    """
    NONE = "None"
    LEGACY_LAYOUT = "Layout1"
    LAYOUT = "Layout2"
    LAYOUT2 = "Layout2"
    CONTENT = "Content"
    STRICT = "Strict"
    EXACT = "Exact"


class ExactMatchSettings(object):
    """
    Encapsulates settings for the "Exact" match level.
    """

    def __init__(self, min_diff_intensity=0, min_diff_width=0, min_diff_height=0, match_threshold=0.0):
        # type: (int, int, int, float) -> None
        """
        Ctor.

        :param min_diff_intensity: Minimal non-ignorable pixel intensity difference.
        :param min_diff_width: Minimal non-ignorable diff region width.
        :param min_diff_height: Minimal non-ignorable diff region height.
        :param match_threshold: The ratio of differing pixels above which images are considered mismatching.
        """
        self.min_diff_intensity = min_diff_intensity  # type: int
        self.min_diff_width = min_diff_width  # type: int
        self.min_diff_height = min_diff_height  # type: int
        self.match_threshold = match_threshold  # type: float

    def __getstate__(self):
        return dict(minDiffIntensity=self.min_diff_intensity,
                    minDiffWidth=self.min_diff_width,
                    minDiffHeight=self.min_diff_height,
                    matchThreshold=self.match_threshold)

    # This is required in order for jsonpickle to work on this object.
    # noinspection PyMethodMayBeStatic
    def __setstate__(self, state):
        raise EyesError('Cannot create ExactMatchSettings instance from dict!')

    def __str__(self):
        return "[min diff intensity: %d, min diff width: %d, min diff height: %d, match threshold: %f]" % (
            self.min_diff_intensity, self.min_diff_width, self.min_diff_height, self.match_threshold)


class ImageMatchSettings(object):
    """
    Encapsulates match settings for the a session.
    """

    def __init__(self, match_level=MatchLevel.STRICT, exact_settings=None):
        # type: (tp.Text, tp.Optional[ExactMatchSettings]) -> None
        """
        :param match_level: The "strictness" level of the match.
        :param exact_settings: Parameter for fine tuning the match when "Exact" match level is used.
        """
        self.match_level = match_level  # type: tp.Text
        self.exact_settings = exact_settings  # type: tp.Optional[ExactMatchSettings]

    def __getstate__(self):
        return dict(matchLevel=self.match_level, exact=self.exact_settings)

    # This is required in order for jsonpickle to work on this object.
    # noinspection PyMethodMayBeStatic
    def __setstate__(self, state):
        raise EyesError('Cannot create ImageMatchSettings instance from dict!')

    def __str__(self):
        return "[Match level: %s, Exact match settings: %s]" % (self.match_level, self.exact_settings)



from __future__ import absolute_import

from .utils import ABC
from . import logger

__all__ = ('FixedScaleProvider', 'NullScaleProvider', 'ContextBasedScaleProvider')


class ScaleProvider(ABC):
    _UNKNOWN_SCALE_RATIO = 0.0

    def __init__(self, *args, **kwargs):
        self._scale_ratio = self._UNKNOWN_SCALE_RATIO
        self.device_pixel_ratio = 1

    @property
    def scale_ratio(self):
        return self._scale_ratio

    def update_scale_ratio(self, image_to_scale_width):
        pass


class FixedScaleProvider(ScaleProvider):
    def __init__(self, scale_ratio):
        super(FixedScaleProvider, self).__init__(scale_ratio)
        self._scale_ratio = scale_ratio


class NullScaleProvider(FixedScaleProvider):
    _UNKNOWN_SCALE_RATIO = 1.0


class ContextBasedScaleProvider(ScaleProvider):
    _ALLOWED_VS_DEVIATION = 1
    _ALLOWED_DCES_DEVIATION = 10

    def __init__(self, top_level_context_entire_size, viewport_size,
                 device_pixel_ratio, is_mobile_device):
        super(ContextBasedScaleProvider, self).__init__()
        self.top_level_context_entire_size = top_level_context_entire_size
        self.viewport_size = viewport_size
        self.device_pixel_ratio = device_pixel_ratio
        self.is_mobile_device = is_mobile_device

    @property
    def scale_ratio(self):
        return self._scale_ratio

    @staticmethod
    def get_scale_ratio_to_viewport(viewport_width, image_to_scale_width, current_scale_ratio):
        scaled_image_width = round(image_to_scale_width * viewport_width)
        from_scaled_to_viewport_ratio = viewport_width / scaled_image_width
        return current_scale_ratio * from_scaled_to_viewport_ratio

    def update_scale_ratio(self, image_to_scale_width):
        viewport_width = self.viewport_size['width']
        dces_width = self.top_level_context_entire_size['width']

        # If the image's width is the same as the viewport's width or the
        # top level context's width, no scaling is necessary.
        allowed_vs_deviation = (viewport_width - self._ALLOWED_VS_DEVIATION <= image_to_scale_width
                                <= viewport_width + self._ALLOWED_VS_DEVIATION)
        allowed_dces_deviation = (dces_width - self._ALLOWED_DCES_DEVIATION <= image_to_scale_width
                                  <= dces_width + self._ALLOWED_DCES_DEVIATION)
        if allowed_dces_deviation or allowed_vs_deviation:
            logger.info('Image is already scaled correctly.')
            self._scale_ratio = 1.0
        else:
            logger.info('Calculating the scale ratio..')
            self._scale_ratio = 1.0 / self.device_pixel_ratio
            if self.is_mobile_device:
                logger.info('Mobile device, so using 2 step calculation for scale ration...')
                logger.info('Scale ratio based on DRP: ' + self._scale_ratio)
                self._scale_ratio = self.get_scale_ratio_to_viewport(viewport_width,
                                                                     image_to_scale_width,
                                                                     self._scale_ratio)
            logger.info("Final scale ratio: {}".format(self.scale_ratio))



import typing as tp

from applitools.core.errors import EyesIllegalArgument
from .compat import iteritems


def get_name_from_param(param):
    for var_name, var in iteritems(locals()):
        if var is param:
            return var_name


def not_none(param):
    """
    Fails if the input parameter is None
    """
    if param is None:
        raise ValueError('{} is None'.format(get_name_from_param(param)))


def not_none_or_empty(param):
    """
    Fails if the input parameter string is null or empty.
    """
    not_none(param)
    if not param:
        raise ValueError('{} is empty'.format(get_name_from_param(param)))


def greater_than_or_equal_to_zero(param, *args):
    # type: (int, *int) -> None
    """
    Fails if the input integer parameter is negative.
    """
    if args:
        # adaptation for attr library
        param = args[1]
    if 0 >= param:
        raise ValueError('{} < 0'.format(get_name_from_param(param)))


# TODO: update after resolving issue
def greater_than_zero(param):
    # type: (int) -> None
    if 0 >= param:
        raise ValueError('{} < 1'.format(get_name_from_param(param)))


def is_valid_state(is_valid, message):
    """
    Fails if is_valid is false.
    """
    if not is_valid:
        raise Exception(message)


def is_a(param, klass):
    if not isinstance(param, klass):
        raise EyesIllegalArgument('{} is not instance of {}'.format(param, klass))



import os
import uuid
import typing as tp
from datetime import datetime

from .errors import EyesError
from .utils import general_utils

__all__ = ('BatchInfo', 'CoordinatesType')


class BatchInfo(object):
    """
    A batch of tests.
    """

    def __init__(self, name=None, started_at=None):
        # type: (tp.Optional[tp.Text], tp.Optional[datetime]) -> None
        if started_at is None:
            started_at = datetime.now(general_utils.UTC)

        self.name = name if name else os.environ.get('APPLITOOLS_BATCH_NAME', None)  # type: tp.Optional[tp.Text]
        self.started_at = started_at  # type: datetime
        self.id = os.environ.get('APPLITOOLS_BATCH_ID', str(uuid.uuid4()))  # type: tp.Text

    @property
    def id_(self):
        # TODO: Remove in this way of initialization in future
        return self.id

    @id_.setter
    def id_(self, value):
        self.id = value

    def __getstate__(self):
        return dict(name=self.name, startedAt=self.started_at.isoformat(), id=self.id)

    # Required is required in order for jsonpickle to work on this object.
    # noinspection PyMethodMayBeStatic
    def __setstate__(self, state):
        raise EyesError('Cannot create BatchInfo instance from dict!')

    def __str__(self):
        return "%s - %s - %s" % (self.name, self.started_at, self.id)


class CoordinatesType(object):
    """
     Encapsulates the type of coordinates used by the region provider.
    """

    # The coordinates should be used "as is" on the screenshot image.
    # Regardless of the current context.
    SCREENSHOT_AS_IS = 'SCREENSHOT_AS_IS'

    # The coordinates should be used "as is" within the current context. For
    # example, if we're inside a frame, the coordinates are "as is",
    # but within the current frame's viewport.
    CONTEXT_AS_IS = 'CONTEXT_AS_IS'

    # Coordinates are relative to the context. For example, if we are in
    # a context of a frame in a web page, then the coordinates are relative to
    # the  frame. In this case, if we want to crop an image region based on
    # an element's region, we will need to calculate their respective "as
    # is" coordinates.
    CONTEXT_RELATIVE = 'CONTEXT_RELATIVE'



from __future__ import absolute_import

import typing as tp

if tp.TYPE_CHECKING:
    from selenium.webdriver.remote.webdriver import WebDriver
    from selenium.webdriver.remote.webelement import WebElement

    from applitools.core.geometry import Region
    from applitools.eyes_selenium.webdriver import EyesWebDriver
    from applitools.eyes_selenium.webelement import EyesWebElement

    RunningSession = tp.Dict[tp.Text, tp.Any]
    ViewPort = tp.Dict[tp.Text, int]
    AppOutput = tp.Dict[tp.Text, tp.Any]
    MatchResult = tp.Dict[tp.Text, tp.Any]
    AppEnvironment = tp.Dict[tp.Text, tp.Any]
    SessionStartInfo = tp.Dict[tp.Text, tp.Any]
    Num = tp.Union[int, float]

    AnyWebDriver = tp.Union[EyesWebDriver, WebDriver]
    AnyWebElement = tp.Union[EyesWebElement, WebElement]
    FrameReference = tp.Union[tp.Text, int, EyesWebElement, WebElement]

    # could contain MouseTrigger, TextTrigger
    UserInputs = tp.List
    RegionOrElement = tp.Union[EyesWebElement, Region]



"""
Contains the Trigger classes (e.g., click, text writing) used by Eyes.
"""
from __future__ import absolute_import

import typing as tp
from collections import OrderedDict

from .errors import EyesError

if tp.TYPE_CHECKING:
    from .geometry import Point, Region

__all__ = ('TextTrigger', 'MouseTrigger')

MOUSE_ACTION = {'click': 'Click', 'right_click': 'RightClick', 'double_click': 'DoubleClick',
                'move': 'Move', 'down': 'Down', 'up': 'Up'}

_TRIGGER_TYPE_FIELD_NAME = "triggerType"
_TRIGGER_TYPE = {'unknown': 'Unknown', 'mouse': 'Mouse', 'text': 'Text', 'keyboard': 'Keyboard'}


class MouseTrigger(object):
    """Encapsulates a mouse trigger."""

    def __init__(self, action, control, location):
        # type: (tp.Text, Region, Point) -> None
        """
        Ctor.

        :param action: The mouse action represented by the trigger.
        :param control: The region to which the text was entered.
        :param location: The location which was clicked relative to the control.
        """
        self.action = action
        self.control = control
        self.location = location  # Point. Relative to the control's (left,top) corner.

    def __getstate__(self):
        return OrderedDict([(_TRIGGER_TYPE_FIELD_NAME, _TRIGGER_TYPE['mouse']),
                            ("mouseAction", MOUSE_ACTION[self.action]), ("control", self.control),
                            ("location", self.location)])

    # Required is required in order for jsonpickle to work on this object.
    # noinspection PyMethodMayBeStatic
    def __setstate__(self, state):
        raise EyesError('Cannot create MouseTrigger instance from dict!')

    def __str__(self):
        return "%s [%s] %s" % (self.action, self.control, self.location)


class TextTrigger(object):
    """Encapsulates a text input by the user."""

    def __init__(self, control, text):
        # type: (Region, tp.Text) -> None
        """
        Ctor.

        :param control: The region to which the text was entered.
        :param text: The trigger's text.
        """
        self.control = control
        self.text = text

    def __getstate__(self):
        return OrderedDict([(_TRIGGER_TYPE_FIELD_NAME, _TRIGGER_TYPE['text']),
                            ("control", self.control), ("text", self.text)])

    # Required is required in order for jsonpickle to work on this object.
    # noinspection PyMethodMayBeStatic
    def __setstate__(self, state):
        raise EyesError('Cannot create MouseTrigger instance from dict!')

    def __str__(self):
        return "Text [%s] %s" % (self.control, self.text)



"""
Compatibility layer between Python 2 and 3
"""
from __future__ import absolute_import

import io
import abc
import sys
from gzip import GzipFile

__all__ = ('ABC', 'range', 'iteritems')

PY3 = sys.version_info >= (3,)

if PY3:
    from urllib.parse import urlparse, urljoin
    from gzip import compress as gzip_compress
    from queue import Queue

    ABC = abc.ABC
    range = range  # type: ignore
else:
    from urlparse import urlparse, urljoin
    from Queue import Queue

    ABC = abc.ABCMeta(str("ABC"), (), {})
    range = xrange  # type: ignore  # noqa: F821

    def gzip_compress(data, compresslevel=9):
        """Compress data in one shot and return the compressed string.
        Optional argument is the compression level, in range of 0-9.
        """
        buf = io.BytesIO()
        with GzipFile(fileobj=buf, mode='wb', compresslevel=compresslevel) as f:
            f.write(data)
        return buf.getvalue()


def iteritems(dct):
    return (getattr(dct, 'iteritems', None) or dct.items)()



"""
General purpose utilities.
"""
import itertools
import json
import time
import types
import typing as tp
from datetime import timedelta, tzinfo

from .. import logger
from .compat import urlparse

if tp.TYPE_CHECKING:
    from selenium.webdrivetr.remote.webdriver import WebDriver
    from selenium.webdriver.remote.webelement import WebElement
    from selenium.webdriver.remote.switch_to import SwitchTo

    from applitools.eyes_selenium.webdriver import EyesWebDriver, _EyesSwitchTo
    from applitools.eyes_selenium.webelement import EyesWebElement


class _UtcTz(tzinfo):
    """
    A UTC timezone class which is tzinfo compliant.
    """
    _ZERO = timedelta(0)

    def utcoffset(self, dt):
        return _UtcTz._ZERO

    def tzname(self, dt):
        return "UTC"

    def dst(self, dt):
        return _UtcTz._ZERO


# Constant representing UTC
UTC = _UtcTz()


def to_json(obj):
    # type: (tp.Dict[str, tp.Any]) -> str
    """
    Returns an object's json representation (defaults to __getstate__ for user defined types).
    """
    return json.dumps(obj, default=lambda o: o.__getstate__(), indent=4)


def create_proxy_property(property_name, target_name, is_settable=False):
    # type: (str, str, bool) -> property
    """
    Returns a property object which forwards "name" to target.

    :param property_name: The name of the property.
    :param target_name: The target to forward to.
    """

    # noinspection PyUnusedLocal
    def _proxy_get(self):
        # type: (tp.Any) -> tp.Dict[str, float]
        return getattr(getattr(self, target_name), property_name)

    # noinspection PyUnusedLocal
    def _proxy_set(self, val):
        return setattr(getattr(self, target_name), property_name, val)

    if not is_settable:
        return property(_proxy_get)
    else:
        return property(_proxy_get, _proxy_set)


def create_forwarded_method(from_,  # type: tp.Union[EyesWebDriver, EyesWebElement, _EyesSwitchTo]
                            to,  # type: tp.Union[WebDriver, WebElement, SwitchTo]
                            func_name,  # type: str
                            ):
    # type: (...) -> tp.Callable
    """
    Returns a method(!) to be set on 'from_', which activates 'func_name' on 'to'.

    :param from_: Source.
    :param to: Destination.
    :param func_name: The name of function to activate.
    :return: Relevant method.
    """

    # noinspection PyUnusedLocal
    def forwarded_method(self_, *args, **kwargs):
        # type: (EyesWebDriver, *tp.Any, **tp.Any) -> tp.Callable
        return getattr(to, func_name)(*args, **kwargs)

    return types.MethodType(forwarded_method, from_)


def create_proxy_interface(from_,  # type: tp.Union[EyesWebDriver, EyesWebElement, _EyesSwitchTo]
                           to,  # type: tp.Union[WebDriver, WebElement, SwitchTo]
                           ignore_list=None,  # type: tp.List[str]
                           override_existing=False,  # type: bool
                           ):
    # type: (...) -> None
    """
    Copies the public interface of the destination object, excluding names in the ignore_list,
    and creates an identical interface in 'eyes_core', which forwards calls to dst.

    :param from_: Source.
    :param to: Destination.
    :param ignore_list: List of names to ignore while copying.
    :param override_existing: If False, attributes already existing in 'eyes_core' will not be overridden.
    """
    if not ignore_list:
        ignore_list = []
    for attr_name in dir(to):
        if not attr_name.startswith('_') and attr_name not in ignore_list:
            if callable(getattr(to, attr_name)):
                if override_existing or not hasattr(from_, attr_name):
                    setattr(from_, attr_name, create_forwarded_method(from_, to, attr_name))


def cached_property(f):
    # type: (tp.Callable) -> tp.Any
    """
    Returns a cached property that is calculated by function f
    """

    def get(self):
        try:
            return self._property_cache[f]
        except AttributeError:
            self._property_cache = {}
            x = self._property_cache[f] = f(self)
            return x
        except KeyError:
            x = self._property_cache[f] = f(self)
            return x

    return property(get)


def is_absolute_url(url):
    return bool(urlparse(url).netloc)


def is_url_with_scheme(url):
    return bool(urlparse(url).scheme)


def timeit(method):
    def timed(*args, **kw):
        ts = time.time()
        result = method(*args, **kw)
        te = time.time()

        if 'log_time' in kw:
            name = kw.get('log_name', method.__name__.upper())
            kw['log_time'][name] = int((te - ts) * 1000)
        else:
            logger.debug('%r  %2.2f ms' % (method.__name__, (te - ts) * 1000))
        return result

    return timed


def retry(delays=(0, 1, 5),
          exception=Exception,
          report=lambda *args: None):
    """
    This is a Python decorator which helps implementing an aspect oriented
    implementation of a retrying of certain steps which might fail sometimes.
    https://code.activestate.com/recipes/580745-retry-decorator-in-python/
    """

    def wrapper(function):
        def wrapped(*args, **kwargs):
            problems = []
            for delay in itertools.chain(delays, [None]):
                try:
                    return function(*args, **kwargs)
                except exception as problem:
                    problems.append(problem)
                    if delay is None:
                        report("retryable failed definitely:", problems)
                        raise
                    else:
                        report("retryable failed:", problem,
                               "-- delaying for %ds" % delay)
                        time.sleep(delay)

        return wrapped

    return wrapper



"""
Utilities for image manipulation.
"""
from __future__ import absolute_import

import base64
import io
import math
import typing as tp

from PIL import Image

from applitools.core.errors import EyesError

if tp.TYPE_CHECKING:
    from applitools.core.geometry import Region

__all__ = ('image_from_file', 'image_from_bytes', 'image_from_base64',
           'scale_image', 'get_base64', 'get_bytes', 'get_image_part')


def image_from_file(f):
    """
    Reads the PNG data from the given file stream and returns a new Image instance.
    """
    return Image.open(f)


def image_from_bytes(png_bytes):
    # type: (bytes) -> Image.Image
    """
    Reads the PNG data from the given png bytes and returns a new Image instance.

    :param png_bytes: Png bytes.
    :return: Image instance.
    """
    return Image.open(io.BytesIO(png_bytes))


def image_from_base64(base64_str):
    # type: (str) -> Image.Image
    """
    Reads the PNG data from the given png bytes and returns a new Image instance.

    :param png_bytes: Png bytes.
    :return: Image instance.
    """
    return Image.open(io.BytesIO(base64.b64decode(base64_str)))


def scale_image(image, scale_ratio):
    # type: (Image.Image, float) -> Image.Image
    if scale_ratio == 1:
        return image

    image_ratio = float(image.height) / float(image.width)
    scale_width = int(math.ceil(image.width * scale_ratio))
    scale_height = int(math.ceil(scale_width * image_ratio))
    image = image.convert('RGBA')
    scaled_image = image.resize((scale_width, scale_height), resample=Image.BICUBIC)
    return scaled_image


def get_device_pixel_ratio(driver):
    return driver.execute_script('return window.devicePixelRatio;')


def get_base64(image):
    # type: (Image.Image) -> str
    """
    Gets the base64 representation of the PNG bytes.

    :return: The base64 representation of the PNG bytes.
    """
    image_bytes_stream = io.BytesIO()
    image.save(image_bytes_stream, format='PNG')
    image64 = base64.b64encode(image_bytes_stream.getvalue()).decode('utf-8')
    image_bytes_stream.close()
    return image64


def get_bytes(image):
    # type: (Image.Image) -> bytes
    """
    Gets the image bytes.

    :return: The image bytes.
    """
    image_bytes_stream = io.BytesIO()
    image.save(image_bytes_stream, format='PNG')
    image_bytes = image_bytes_stream.getvalue()
    image_bytes_stream.close()
    return image_bytes


def get_image_part(image, region):
    # type: (Image.Image, Region) -> Image.Image
    """
    Get a copy of the part of the image given by region.

    :return: The part of the image.
    """
    if region.is_empty():
        raise EyesError('region is empty!')
    return image.crop(box=(region.left, region.top, region.right, region.bottom))



from __future__ import absolute_import

import typing as tp

__all__ = ('TestResults', 'TestResultsStatus')


class TestResultsStatus(object):
    """
    Status values for tests results.
    """
    Passed = "Passed"
    Unresolved = "Unresolved"
    Failed = "Failed"

    @classmethod
    def get_status(cls, status):
        # type: (tp.Optional[tp.Text]) -> tp.Text
        status = status.lower() if status is not None else ''
        if status == cls.Passed.lower():
            return cls.Passed
        elif status == cls.Unresolved.lower():
            return cls.Unresolved
        elif status == cls.Failed.lower():
            return cls.Failed
        else:
            # Unknown status
            return status


class TestResults(object):
    """
    Eyes test results.

    # TODO: update regarding JAVA SDK
    """

    def __init__(self, steps=0,  # type: int
                 matches=0,  # type: int
                 mismatches=0,  # type: int
                 missing=0,  # type: int
                 exact_matches=0,  # type: int
                 strict_matches=0,  # type: int
                 content_matches=0,  # type: int
                 layout_matches=0,  # type: int
                 none_matches=0,  # type: int
                 status=None  # type: tp.Optional[tp.Text]
                 ):
        # type: (...) -> None
        self.steps = steps
        self.matches = matches
        self.mismatches = mismatches
        self.missing = missing
        self.exact_matches = exact_matches
        self.strict_matches = strict_matches
        self.content_matches = content_matches
        self.layout_matches = layout_matches
        self.none_matches = none_matches
        self._status = status
        self.is_new = None  # type: tp.Optional[bool]
        self.url = None  # type: tp.Optional[tp.Text]

    @property
    def status(self):
        # type: () -> tp.Text
        return TestResultsStatus.get_status(self._status)

    @status.setter
    def status(self, value):
        # type: (tp.Text) -> None
        self._status = value

    @property
    def is_passed(self):
        # type: () -> bool
        return (self.status is not None) and self.status.lower() == TestResultsStatus.Passed.lower()

    def _to_dict(self):
        return dict(steps=self.steps, matches=self.matches, mismatches=self.mismatches,
                    missing=self.missing, exact_matches=self.exact_matches,
                    strict_matches=self.strict_matches, content_matches=self.content_matches,
                    layout_matches=self.layout_matches, none_matches=self.none_matches,
                    is_new=self.is_new, url=self.url, status=self.status)

    def __str__(self):
        if self.is_new is not None:
            is_new = "New test" if self.is_new else "Existing test"
        else:
            is_new = ""
        return "%s [ steps: %d, matches: %d, mismatches: %d, missing: %d ], URL: %s" % \
               (is_new, self.steps, self.matches, self.mismatches, self.missing, self.url)



__all__ = ('EyesError', 'EyesIllegalArgument', 'OutOfBoundsError', 'TestFailedError', 'NewTestError',
           'DiffsFoundError', 'CoordinatesTypeConversionError')


class EyesError(Exception):
    """
    Applitools Eyes Exception.
    """


class EyesIllegalArgument(EyesError):
    """
    Raise when parameter with wrong type passed to function
    """


class OutOfBoundsError(EyesError):
    """
    Indicates that an element is outside a specific boundary (e.g, region outside a frame,
    or point outside an image).
    """


class TestFailedError(Exception):
    """
    Indicates that a test did not pass (i.e., test either failed or is a new test).
    """

    def __init__(self, message, test_results=None):
        self.message = message
        self.test_results = test_results

    def __str__(self):
        return "%s , %s" % (self.message, self.test_results)


class NewTestError(TestFailedError):
    """
    Indicates that a test is a new test.
    """

    def __init__(self, message, test_results=None):
        super(NewTestError, self).__init__(message, test_results)


class DiffsFoundError(TestFailedError):
    """
    Indicates that an existing test ended, and that differences where found from the baseline.
    """

    def __init__(self, message, test_results=None):
        super(DiffsFoundError, self).__init__(message, test_results)


class CoordinatesTypeConversionError(EyesError):
    def __init__(self, from_, to, test_results=None):
        message = 'Cannot convert from {} to {}'.format(from_, to)
        super(CoordinatesTypeConversionError, self).__init__(message, test_results)



import typing as tp

import attr

from applitools.core.geometry import Region
from applitools.core.utils import ABC

if tp.TYPE_CHECKING:
    from applitools.core.capture import EyesScreenshot

__all__ = ('RegionProvider', 'NullRegionProvider', 'NULL_REGION_INSTANCE')


@attr.s
class RegionProvider(ABC):
    """
    Encapsulates a getRegion "callback" and how the region's coordinates should be used.
    """
    _region = attr.ib()

    def get_region(self, eyes_screenshot=None):
        # type: (tp.Optional[EyesScreenshot]) -> Region
        """
        :return: A region with "as is" viewport coordinates.
        """
        return self._region

    def __str__(self):
        return str(self._region)


@attr.s
class NullRegionProvider(RegionProvider):
    _region = attr.ib(init=False, factory=Region.create_empty_region)


NULL_REGION_INSTANCE = NullRegionProvider()



import abc
import typing as tp

import attr

from applitools.core.geometry import Point
from applitools.core.utils.compat import ABC

if tp.TYPE_CHECKING:
    from applitools.core.utils.custom_types import ViewPort

__all__ = ('PositionProvider', 'InvalidPositionProvider', 'PositionMemento')


@attr.s
class PositionProvider(ABC):
    _states = attr.ib(init=False, factory=list)

    @abc.abstractmethod
    def get_current_position(self):
        # type: () -> tp.Optional[Point]
        """
        :return: The current position, or `None` if position is not available.
        """

    @abc.abstractmethod
    def set_position(self, location):
        # type: (Point) -> None
        """
        Go to the specified location.

        :param location: The position to set
        """

    @property
    @abc.abstractmethod
    def get_entire_size(self):
        # type: () -> ViewPort
        """
        :return: The entire size of the container which the position is relative to.
        """

    def push_state(self):
        """
        Adds the current position to the states list.
        """
        self._states.append(self.get_current_position())

    def pop_state(self):
        """
        Sets the position to be the last position added to the states list.
        """
        self.set_position(self._states.pop())


class InvalidPositionProvider(PositionProvider):
    """
    An implementation of :py:class:`PositionProvider` which throws an exception
    for every method. Can be used as a placeholder until an actual
    implementation is set.
    """

    def get_current_position(self):
        raise NotImplementedError("This class does not implement methods!")

    def set_position(self, location):
        # type: (Point) -> None
        raise NotImplementedError("This class does not implement methods!")

    def get_entire_size(self):
        raise NotImplementedError("This class does not implement methods!")


class PositionMemento(ABC):
    """
    A base class for position related memento instances. This is intentionally
    not an interface, since the mementos might vary in their interfaces.
    """



from . import image_utils, argument_guard
from .general_utils import cached_property
from .compat import ABC, range, iteritems

__all__ = (compat.__all__ +  # noqa
           ('image_utils', 'argument_guard')
           )



import re
import codecs
import sys
from os import path

from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))


def read(filename):
    return codecs.open(path.join(here, filename), 'r', 'utf-8').read()


# preventing ModuleNotFoundError caused by importing lib before installing deps
def get_version(package_name):
    version_file = read('applitools/{}/__version__.py'.format(package_name))
    try:
        version = re.findall(r"^__version__ = '([^']+)'\r?$",
                             version_file, re.M)[0]
    except IndexError:
        raise RuntimeError('Unable to determine version.')

    return version


install_requires = [
    'attrs>=18',
    'Pillow>=5.0.0',
    'requests>=2.1.0',
    # uses for check if the mobile platform is used in the test
    'Appium-Python-Client>=0.31'
]
# using this way of defining instead of 'typing>=3.5.2; python_version<="3.4"'
# for run on old version of setuptools without issues
if sys.version_info < (3, 5):
    # typing module was added as builtin in Python 3.5
    install_requires.append('typing >= 3.5.2')

setup(
    name='eyes_core',
    version=get_version('core'),
    packages=find_packages(include=['applitools.*'], exclude=('tests',)),
    url='http://www.applitools.com',
    license='Apache License, Version 2.0',
    author='Applitools Team',
    author_email='team@applitools.com',
    description='Applitools Python SDK. Core package',
    long_description=read('README.rst'),
    long_description_content_type='text/x-rst',
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Topic :: Software Development :: Quality Assurance",
        "Topic :: Software Development :: Testing"
    ],
    keywords='applitools eyes eyes_core',
    install_requires=install_requires,
    package_data={
        '':     ['README.rst', 'LICENSE'],
        'core': ['py.typed'],
    },
    project_urls={
        'Bug Reports': 'https://github.com/applitools/eyes.sdk.python/issues',
        'Source':      'https://github.com/applitools/eyes.sdk.python/tree/master/eyes_core/applitools/core',
    },
)



from .position_provider import PositionProvider, InvalidPositionProvider, PositionMemento
from .region_provider import RegionProvider, NullRegionProvider, NULL_REGION_INSTANCE

__all__ = (
    position_provider.__all__ +  # noqa
    region_provider.__all__  # noqa
)



import abc
import typing as tp

import attr

from .geometry import Point, Region
from .metadata import CoordinatesType
from .utils import ABC, image_utils

if tp.TYPE_CHECKING:
    from PIL import Image

__all__ = ('EyesScreenshot',)


@attr.s
class EyesScreenshot(ABC):
    """
     Base class for handling screenshots.
     """

    _image = attr.ib()  # type: Image.Image

    @abc.abstractmethod
    def sub_screenshot(self, region, throw_if_clipped=False):
        # type: (Region, bool) -> Region
        pass

    @abc.abstractmethod
    def convert_location(self, location, from_, to):
        # type: (Point, CoordinatesType, CoordinatesType) -> Point
        pass

    @abc.abstractmethod
    def location_in_screenshot(self, location, coordinates_type):
        # type: (Point, CoordinatesType) -> Point
        pass

    @abc.abstractmethod
    def intersected_region(self, region, result_coordinate_types):
        # type: (Region, CoordinatesType) -> Region
        pass

    def convert_region_location(self, region, from_, to):
        # type: (Region, CoordinatesType, CoordinatesType) -> Region
        assert region is not None
        assert isinstance(region, Region)
        assert from_ is not None
        assert to is not None

        if region.is_empty:
            return Region.create_empty_region()

        updated_location = self.convert_location(region.location, from_, to)
        return Region(updated_location.x, updated_location.y, region.width, region.height)

    @property
    def image_region(self):
        # type: () -> Region
        return Region(0, 0, self._image.width, self._image.height, CoordinatesType.SCREENSHOT_AS_IS)

    @staticmethod
    def from_region(region):
        # type: (Region) -> Image.Image
        return Image.new('RGBA', (region.width, region.height))

    def get_bytes(self):
        # type: () -> bytes
        """
        Returns the bytes of the screenshot.

        :return: The bytes representation of the png.
        """
        return image_utils.get_bytes(self._image)



from .triggers import *  # noqa
from .test_results import *  # noqa
from .match_window_task import *  # noqa
from .logger import *  # noqa
from .errors import *  # noqa
from .capture import *  # noqa
from .scaling import *  # noqa
from .eyes_base import *  # noqa
from .geometry import *  # noqa
from .agent_connector import AgentConnector  # noqa
from .positioning import *  # noqa
from .match import *  # noqa
from .metadata import *  # noqa

__all__ = (triggers.__all__ +  # noqa
           test_results.__all__ +  # noqa
           match_window_task.__all__ +  # noqa
           logger.__all__ +  # noqa
           errors.__all__ +  # noqa
           scaling.__all__ +  # noqa
           capture.__all__ +  # noqa
           eyes_base.__all__ +  # noqa
           geometry.__all__ +  # noqa
           positioning.__all__ +  # noqa
           match.__all__ +  # noqa
           metadata.__all__ +  # noqa
           ('logger', 'AgentConnector'))



from __future__ import absolute_import

import abc
import os
import typing as tp

from . import logger
from .utils import ABC
from .match import ImageMatchSettings
from .metadata import BatchInfo
from .agent_connector import AgentConnector
from .errors import EyesError, NewTestError, DiffsFoundError, TestFailedError
from .match_window_task import MatchWindowTask
from .test_results import TestResults, TestResultsStatus

if tp.TYPE_CHECKING:
    from .utils.custom_types import (ViewPort, UserInputs, AppEnvironment, MatchResult,
                                     RunningSession, SessionStartInfo, RegionOrElement)
    from .capture import EyesScreenshot

__all__ = ('FailureReports', 'EyesBase')


class FailureReports(object):
    """
    Failures are either reported immediately when they are detected, or when the test is closed.
    """
    IMMEDIATE = "Immediate"
    ON_CLOSE = "OnClose"


class EyesBase(ABC):
    _DEFAULT_MATCH_TIMEOUT = 2000  # Milliseconds
    _DEFAULT_WAIT_BEFORE_SCREENSHOTS = 100  # ms
    DEFAULT_EYES_SERVER = 'https://eyessdk.applitools.com'

    def __init__(self, server_url=DEFAULT_EYES_SERVER):
        # type: (tp.Text) -> None
        """
        Creates a new (possibly disabled) Eyes instance that interacts with the Eyes server.

        :param server_url: The URL of the Eyes server
        """
        self._agent_connector = AgentConnector(server_url)  # type: AgentConnector
        self._should_get_title = False  # type: bool
        self._is_open = False  # type: bool
        self._app_name = None  # type: tp.Optional[tp.Text]
        self._running_session = None  # type: tp.Optional[RunningSession]
        self._match_timeout = EyesBase._DEFAULT_MATCH_TIMEOUT  # type: int
        self._last_screenshot = None  # type: tp.Optional[EyesScreenshot]
        self._should_match_once_on_timeout = False  # type: bool
        self._start_info = None  # type: tp.Optional[SessionStartInfo]
        self._test_name = None  # type: tp.Optional[tp.Text]
        self._user_inputs = []  # type: UserInputs
        self._region_to_check = None  # type: tp.Optional[RegionOrElement]
        self._viewport_size = None  # type: ViewPort

        # key-value pairs to be associated with the test. Can be used for filtering later.
        self._properties = []  # type: tp.List

        # Disables Applitools Eyes and uses the webdriver directly.
        self.is_disabled = False  # type: bool

        # An optional string identifying the current library using the SDK.
        self.agent_id = None  # type: tp.Optional[tp.Text]

        # Should the test report mismatches immediately or when it is finished. See FailureReports.
        self.failure_reports = FailureReports.ON_CLOSE  # type: tp.Text

        # The default match settings for the session. See ImageMatchSettings.
        self.default_match_settings = ImageMatchSettings()  # type: ImageMatchSettings

        # The batch to which the tests belong to. See BatchInfo. None means no batch.
        self.batch = None  # type: tp.Optional[BatchInfo]

        # A string identifying the OS running the AUT. Use this to override Eyes automatic inference.
        self.host_os = None  # type: tp.Optional[tp.Text]

        # A string identifying the app running the AUT. Use this to override Eyes automatic inference.
        self.host_app = None  # type: tp.Optional[tp.Text]

        # A string that, if specified, determines the baseline to compare with and disables automatic baseline
        # inference.
        self.baseline_branch_name = None  # type: tp.Optional[tp.Text]

        # A boolean denoting whether new tests should be automatically accepted.
        self.save_new_tests = True  # type: bool

        # Whether failed tests should be automatically saved with all new output accepted.
        self.save_failed_tests = False  # type: bool

        # A string identifying the branch in which tests are run.
        self.branch_name = None  # type: tp.Optional[tp.Text]

        # A string identifying the parent branch of the branch set by "branch_name".
        self.parent_branch_name = None  # type: tp.Optional[tp.Text]

        # If true, Eyes will treat new tests the same as failed tests.
        self.fail_on_new_test = False  # type: bool

        # The number of milliseconds to wait before each time a screenshot is taken.
        self.wait_before_screenshots = EyesBase._DEFAULT_WAIT_BEFORE_SCREENSHOTS  # type: int

        # If true, we will send full DOM to the server for analyzing
        self.send_dom = False  # type: bool

    @property
    @abc.abstractmethod
    def _title(self):
        # type: () -> tp.Text
        """
        Returns the title of the window of the AUT, or empty string if the title is not available.
        """

    @abc.abstractmethod
    def get_screenshot(self, **kwargs):
        pass

    @abc.abstractmethod
    def get_viewport_size(self):
        # type: () -> ViewPort
        """
        :return: The viewport size of the AUT.
        """

    @abc.abstractmethod
    def set_viewport_size(self, size):
        # type: (ViewPort) -> None
        """
        :param size: The required viewport size.
        """

    @abc.abstractmethod
    def _assign_viewport_size(self):
        # type: () -> None
        """
        Assign the viewport size we need to be in the default content frame.
        """

    @abc.abstractmethod
    def _environment(self):
        # type: () -> AppEnvironment
        """
        Application environment is the environment (e.g., the host OS) which runs the application under test.

        :return: The current application environment.
        """

    @abc.abstractmethod
    def _inferred_environment(self):
        pass

    @property
    def _seconds_to_wait_screenshot(self):
        return self.wait_before_screenshots / 1000.0

    @property
    def match_level(self):
        # type: () -> tp.Text
        """
        Gets the default match level for the entire session. See ImageMatchSettings.
        """
        return self.default_match_settings.match_level

    @match_level.setter
    def match_level(self, match_level):
        # type: (tp.Text) -> None
        """
        Sets the default match level for the entire session. See ImageMatchSettings.

        :param match_level: The match level to set. Should be one of the values defined by MatchLevel
        """
        self.default_match_settings.match_level = match_level

    @property
    def match_timeout(self):
        # type: () -> int
        """
        Gets the default timeout for check_XXXX operations. (milliseconds)

        :return: The match timeout (milliseconds)
        """
        return self._match_timeout

    @match_timeout.setter
    def match_timeout(self, match_timeout):
        # type: (int) -> None
        """
        Sets the default timeout for check_XXXX operations. (milliseconds)
        """
        if 0 < match_timeout < MatchWindowTask.MINIMUM_MATCH_TIMEOUT:
            raise ValueError("Match timeout must be at least 60ms.")
        self._match_timeout = match_timeout

    @property
    def api_key(self):
        # type: () -> tp.Text
        """
        Gets the Api key used for authenticating the user with Eyes.

        :return: The Api key used for authenticating the user with Eyes.
        """
        return self._agent_connector.api_key

    @api_key.setter
    def api_key(self, api_key):
        # type: (tp.Text) -> None
        """
        Sets the api key used for authenticating the user with Eyes.

        :param api_key: The api key used for authenticating the user with Eyes.
        """
        self._agent_connector.api_key = api_key  # type: ignore

    @property
    def server_url(self):
        # type: () -> tp.Text
        """
        Gets the URL of the Eyes server.

        :return: The URL of the Eyes server, or None to use the default server.
        """
        return self._agent_connector.server_url

    @server_url.setter
    def server_url(self, server_url):
        # type: (tp.Text) -> None
        """
        Sets the URL of the Eyes server.

        :param server_url: The URL of the Eyes server, or None to use the default server.
        :return: None
        """
        if server_url is None:
            self._agent_connector.server_url = EyesBase.DEFAULT_EYES_SERVER
        else:
            self._agent_connector.server_url = server_url

    @property
    @abc.abstractmethod
    def base_agent_id(self):
        # type: () -> tp.Text
        """
        Must return version of SDK. (e.g. Selenium, Images) in next format:
            "eyes.{package}.python/{lib_version}"
        """

    @property
    def full_agent_id(self):
        # type: () -> tp.Text
        """
        Gets the agent id, which identifies the current library using the SDK.

        :return: The agent id.
        """
        if self.agent_id is None:
            return self.base_agent_id
        return "{0} [{1}]".format(self.agent_id, self.base_agent_id)

    def add_property(self, name, value):
        # type: (tp.Text, tp.Text) -> None
        """
        Associates a key/value pair with the test. This can be used later for filtering.
        :param name: (string) The property name.
        :param value: (string) The property value
        """
        self._properties.append({'name': name, 'value': value})

    @property
    def is_open(self):
        # type: () -> bool
        """
        Returns whether the session is currently running.
        """
        return self._is_open

    def close(self, raise_ex=True):
        # type: (bool) -> tp.Optional[TestResults]
        """
        Ends the test.

        :param raise_ex: If true, an exception will be raised for failed/new tests.
        :return: The test results.
        """
        if self.is_disabled:
            logger.debug('close(): ignored (disabled)')
            return None
        try:
            logger.debug('close({})'.format(raise_ex))
            if not self._is_open:
                raise ValueError("Eyes not open")

            self._is_open = False

            self._reset_last_screenshot()

            # If there's no running session, we simply return the default test results.
            if not self._running_session:
                logger.debug('close(): Server session was not started')
                logger.info('close(): --- Empty test ended.')
                return TestResults()

            is_new_session = self._running_session['is_new_session']
            results_url = self._running_session['session_url']

            logger.info("close(): Ending server session...")
            should_save = (is_new_session and self.save_new_tests) or \
                          ((not is_new_session) and self.save_failed_tests)
            logger.debug("close(): automatically save session? %s" % should_save)
            results = self._agent_connector.stop_session(self._running_session, False, should_save)
            results.is_new = is_new_session
            results.url = results_url
            logger.info("close(): %s" % results)

            if results.status == TestResultsStatus.Unresolved:
                if results.is_new:
                    instructions = "Please approve the new baseline at " + results_url
                    logger.info("--- New test ended. " + instructions)
                    if raise_ex:
                        message = "'%s' of '%s'. %s" % (self._start_info['scenarioIdOrName'],
                                                        self._start_info['appIdOrName'],
                                                        instructions)
                        raise NewTestError(message, results)
                else:
                    logger.info("--- Failed test ended. See details at {}".format(results_url))
                    if raise_ex:
                        raise DiffsFoundError("Test '{}' of '{}' detected differences! See details at: {}".format(
                            self._start_info['scenarioIdOrName'],
                            self._start_info['appIdOrName'],
                            results_url), results)
            elif results.status == TestResultsStatus.Failed:
                logger.info("--- Failed test ended. See details at {}".format(results_url))
                if raise_ex:
                    raise TestFailedError("Test '{}' of '{}'. See details at: {}".format(
                        self._start_info['scenarioIdOrName'],
                        self._start_info['appIdOrName'],
                        results_url), results)
            # Test passed
            logger.info("--- Test passed. See details at {}".format(results_url))

            return results
        finally:
            self._running_session = None
            logger.close()

    def abort_if_not_closed(self):
        # type: () -> None
        """
        If a test is running, aborts it. Otherwise, does nothing.
        """
        if self.is_disabled:
            logger.debug('abort_if_not_closed(): ignored (disabled)')
            return
        try:
            self._reset_last_screenshot()

            if self._running_session:
                logger.debug('abort_if_not_closed(): Aborting session...')
                try:
                    self._agent_connector.stop_session(self._running_session, True, False)
                    logger.info('--- Test aborted.')
                except EyesError as e:
                    logger.info("Failed to abort server session: %s " % e)
                    pass
                finally:
                    self._running_session = None
        finally:
            logger.close()

    def _before_open(self):
        pass

    def _after_open(self):
        pass

    def _open_base(self, app_name, test_name, viewport_size=None):
        # type: (tp.Text, tp.Text, tp.Optional[ViewPort]) -> None
        """
        Starts a test.

        :param app_name: The name of the application under test.
        :param test_name: The test name.
        :param viewport_size: The client's viewport size (i.e., the visible part of the document's body) or None to
                                allow any viewport size.
        :return: An updated web driver
        :raise EyesError: If the session was already open.
        """
        logger.open_()
        if self.is_disabled:
            logger.debug('_open_base(): ignored (disabled)')
            return
        logger.info('\nEyes version: {}\n'.format(self.full_agent_id))

        if self.api_key is None:
            try:
                self.api_key = os.environ['APPLITOOLS_API_KEY']
            except KeyError:
                raise EyesError("API key not set! Log in to https://applitools.com to obtain your"
                                " API Key and use 'api_key' to set it.")

        logger.info("open(%s, %s, %s, %s)" % (app_name, test_name, viewport_size, self.failure_reports))

        if self.is_open:
            self.abort_if_not_closed()
            raise EyesError('a test is already running')

        self._before_open()

        self._app_name = app_name
        self._test_name = test_name
        self._viewport_size = viewport_size

        if viewport_size is not None:
            self._ensure_running_session()

        self._is_open = True

        self._after_open()

    def _create_start_info(self):
        # type: () -> None
        app_env = self._environment
        self._start_info = {'agentId':              self.full_agent_id, 'appIdOrName': self._app_name,
                            'scenarioIdOrName':     self._test_name, 'batchInfo': self.batch,
                            'envName':              self.baseline_branch_name, 'environment': app_env,
                            'defaultMatchSettings': self.default_match_settings, 'verId': None,
                            'branchName':           self.branch_name, 'parentBranchName': self.parent_branch_name,
                            'properties':           self._properties}

    def _start_session(self):
        # type: () -> None
        logger.debug("_start_session()")
        self._assign_viewport_size()

        # initialization of Eyes parameters if empty from ENV variables
        if not self.branch_name:
            self.branch_name = os.environ.get('APPLITOOLS_BRANCH', None)
        if not self.baseline_branch_name:
            self.baseline_branch_name = os.environ.get('APPLITOOLS_BASELINE_BRANCH', None)
        if not self.parent_branch_name:
            self.parent_branch_name = os.environ.get('APPLITOOLS_PARENT_BRANCH', None)
        if not self.batch:
            self.batch = BatchInfo()

        self._create_start_info()
        # Actually start the session.
        self._running_session = self._agent_connector.start_session(self._start_info)
        self._should_match_once_on_timeout = self._running_session['is_new_session']

    def _reset_last_screenshot(self):
        # type: () -> None
        self._last_screenshot = None
        self._user_inputs = []  # type: UserInputs

    def _ensure_running_session(self):
        if self._running_session:
            logger.debug('Session already running.')
            return
        self._start_session()
        self._match_window_task = MatchWindowTask(self, self._agent_connector,
                                                  self._running_session,
                                                  self.match_timeout)

    def _before_match_window(self):
        """
        Allow to add custom behavior after receiving response from the server
        """

    def _after_match_window(self):
        """
        Allow to add custom behavior before sending data to the server
        """

    def _check_window_base(self, tag=None, match_timeout=-1, target=None, ignore_mismatch=False):
        if self.is_disabled:
            logger.info("check_window(%s): ignored (disabled)" % tag)
            # TODO: create propper MatchResult class
            result = {"as_expected": True, "screenshot": None}
            return result

        self._ensure_running_session()

        self._before_match_window()

        # TODO: implement MatchWIndow_ analog
        result = self._match_window_task.match_window(retry_timeout=match_timeout,
                                                      tag=tag,
                                                      user_inputs=self._user_inputs,
                                                      default_match_settings=self.default_match_settings,
                                                      target=target,
                                                      ignore_mismatch=ignore_mismatch,
                                                      run_once_after_wait=self._should_match_once_on_timeout)
        self._after_match_window()
        self._handle_match_result(result, tag)
        return result

    def _handle_match_result(self, result, tag):
        # type: (MatchResult, tp.Text) -> None
        self._last_screenshot = result['screenshot']
        as_expected = result['as_expected']
        self._user_inputs = []
        if not as_expected:
            self._should_match_once_on_timeout = True
            if self._running_session and not self._running_session['is_new_session']:
                logger.info("Window mismatch %s" % tag)
                if self.failure_reports == FailureReports.IMMEDIATE:
                    raise TestFailedError("Mismatch found in '%s' of '%s'" %
                                          (self._start_info['scenarioIdOrName'],
                                           self._start_info['appIdOrName']))

    @abc.abstractmethod
    def _try_capture_dom(self):
        # type: () -> tp.Text
        """
        Returns the string with DOM of the current page in the prepared format or empty string
        """

    def _try_post_dom_snapshot(self, dom_json):
        # type: (tp.Text) -> tp.Optional[tp.Text]
        """
        In case DOM data is valid uploads it to the server and return URL where it stored.
        """
        if dom_json is None:
            return None
        try:
            return self._agent_connector.post_dom_snapshot(dom_json)
        except Exception as e:
            logger.warning("Couldn't send DOM Json. Passing...\n Got next error: {}".format(e))
            return None
